@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2015 GraphicsMagick Group
+% Copyright (C) 2003-2019 GraphicsMagick Group
 %
 % This program is covered by multiple licenses, which are described in
 % Copyright.txt. You should have received a copy of Copyright.txt with this
@@ -130,7 +130,7 @@ typedef struct _CineonImageInfo
   U8    sense;                     /* Image sense: 0=positive, 1=negative */
   U32   eol_pad;                   /* End of line padding */
   U32   eoc_pad;                   /* End of channel padding */
-  ASCII reserved2[20];             /* Reserved for future use */  
+  ASCII reserved2[20];             /* Reserved for future use */
 } CineonImageInfo;
 
 typedef struct _CineonImageOriginationInfo
@@ -256,7 +256,7 @@ static unsigned int IsCINEON(const unsigned char *magick,const size_t length)
 \
   if (!IS_UNDEFINED_ASCII(member)) \
     { \
-      (void) strncpy(buffer_,member,Min(sizeof(member),MaxTextExtent)); \
+      (void) memcpy(buffer_,member,Min(sizeof(member),MaxTextExtent)); \
       buffer_[Min(sizeof(member),MaxTextExtent-1)]='\0';             \
       (void) SetImageAttribute(image,name,buffer_); \
       LogSetImageAttribute(name,buffer_); \
@@ -348,7 +348,7 @@ static void SwabCineonImageInfo(CineonImageInfo *image_info)
   MagickSwabFloat(&image_info->blue_primary_chromaticity[0].f);
   MagickSwabFloat(&image_info->blue_primary_chromaticity[1].f);
   MagickSwabUInt32(&image_info->eol_pad);
-  MagickSwabUInt32(&image_info->eoc_pad);  
+  MagickSwabUInt32(&image_info->eoc_pad);
 }
 
 static void SwabCineonImageOriginationInfo(CineonImageOriginationInfo *image_info)
@@ -459,10 +459,10 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
   Image
     *image;
 
-  long
+  unsigned long
     y;
 
-  register long
+  register unsigned long
     x;
 
   register PixelPacket
@@ -472,14 +472,14 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
     i;
 
   size_t
-    offset;    
+    offset;
 
   unsigned int
     channel,
     max_bits_per_sample,
     max_lines_per_image,
     max_pixels_per_line,
-    number_of_channels;    
+    number_of_channels;
 
   MagickPassFail
     status;
@@ -492,7 +492,7 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
 
   unsigned char
     *scandata;
-  
+
   void
     *scanline;
 
@@ -578,7 +578,7 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
   /*
     Obtain offset to pixels.
   */
-  pixels_offset=cin_file_info.image_data_offset;
+  pixels_offset=cin_file_info.image_data_offset & 0xffffffff;
   if (pixels_offset < 712)
     ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
 
@@ -614,7 +614,7 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
       max_lines_per_image=Max(max_lines_per_image,
                               cin_image_info.channel_info[channel].lines_per_image);
     }
-  
+
   image->depth=max_bits_per_sample;
   image->columns=max_pixels_per_line;
   image->rows=max_lines_per_image;
@@ -705,14 +705,14 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
     {
       unsigned char
         *user_data;
-      
+
       const size_t
         block_size = 65536UL;
-      
+
       size_t
         read_size,
         user_data_length;
-      
+
       user_data_length=0UL;
       user_data=(unsigned char *) NULL;
       while (user_data_length < cin_file_info.user_defined_length)
@@ -753,6 +753,29 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
     if (ReadBlobByte(image) == EOF)
       ThrowReaderException(CorruptImageError,UnexpectedEndOfFile,image);
 
+  /*
+    Verify that remaining data is sufficient for claimed dimensions.
+  */
+  if (BlobIsSeekable(image))
+    {
+      magick_off_t file_size;
+      magick_int64_t expected_file_size;
+
+      file_size = GetBlobSize(image);
+
+      expected_file_size=pixels_offset+
+        ((((((magick_int64_t) image->rows*image->columns*number_of_channels+2UL)/3UL)
+           *sizeof(U32)*8UL)+31UL)/32UL)*sizeof(U32);
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "Expected file size %" MAGICK_UINT64_F "u bytes"
+                              " (have %" MAGICK_OFF_F "d bytes)",
+                              expected_file_size, file_size);
+
+      if ((file_size > 0) && ((magick_int64_t) file_size < expected_file_size))
+        ThrowReaderException(CorruptImageError,InsufficientImageDataInFile,image);
+    }
+
   if (image->logging)
     (void) LogMagickEvent(CoderEvent,GetMagickModule(),
                           "Reading Cineon pixels starting at offset %ld",(long) TellBlob(image));
@@ -774,9 +797,11 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
           scandata_bytes=4;
           scale_to_short=64;
           scandata=MagickAllocateMemory(unsigned char *,scandata_bytes);
+          if (scandata == (unsigned char *) NULL)
+            ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
           scanline=scandata;
           MagickBitStreamInitializeRead(&bit_stream,scanline);
-          for (y=0; y < (long) image->rows; y++)
+          for (y=0; y < image->rows; y++)
             {
               q=SetImagePixels(image,0,y,image->columns,1);
               if (q == (PixelPacket *) NULL)
@@ -786,7 +811,7 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
               */
               scanline=scandata;
               i=3;
-              for (x=(long) image->columns; x > 0; x--, i++)
+              for (x=0; x < image->columns; x++, i++)
                 {
                   if (i > 2)
                     {
@@ -802,13 +827,15 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
                   q->opacity=0U;
                   q++;
                 }
+              if (x < image->columns)
+                break;
               if (!SyncImagePixels(image))
                 break;
               if (image->previous == (Image *) NULL)
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
             }
           MagickFreeMemory(scandata);
@@ -816,10 +843,12 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
         }
       case 3:
         {
-          scandata_bytes=image->columns*4;
+          scandata_bytes=MagickArraySize(image->columns,4);
           scale_to_short=64;
           scandata=MagickAllocateMemory(unsigned char *,scandata_bytes);
-          for (y=0; y < (long) image->rows; y++)
+          if (scandata == (unsigned char *) NULL)
+            ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+          for (y=0; y < image->rows; y++)
             {
               magick_uint32_t red;
               magick_uint32_t green;
@@ -832,7 +861,7 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
               if (ReadBlobZC(image,scandata_bytes,&scanline) != scandata_bytes)
                 break;
               MagickBitStreamInitializeRead(&bit_stream,scanline);
-              for (x=0 ; x < (long) image->columns; x++)
+              for (x=0 ; x < image->columns; x++)
                 {
                   /*
                     Packed 10 bit samples with 2 bit pad at end of 32-bit word.
@@ -851,13 +880,15 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
 /*                          (unsigned int)q->red, (unsigned int)q->green, (unsigned int)q->blue); */
                   q++;
                 }
+              if (x < image->columns)
+                break;
               if (!SyncImagePixels(image))
                 break;
               if (image->previous == (Image *) NULL)
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
             }
           MagickFreeMemory(scandata);
@@ -867,6 +898,9 @@ static Image *ReadCINEONImage(const ImageInfo *image_info,
         ThrowReaderException(CorruptImageError,ImageTypeNotSupported,image);
       }
   }
+  if (image->logging)
+    (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                          "EOF at offset %" MAGICK_OFF_F "d", TellBlob(image));
   image->depth=Min(image->depth,QuantumDepth);
   image->colorspace=CineonLogRGBColorspace;
 
@@ -995,7 +1029,7 @@ ModuleExport void UnregisterCINEONImage(void)
   const char \
     *definition_value; \
 \
-  if ((definition_value=AccessDefinition(image_info,"dpx",key+4))) \
+  if ((definition_value=AccessDefinition(image_info,"dpx",&key[4]))) \
     member=(U8) strtol(definition_value, (char **) NULL, 10); \
   else if ((attribute=GetImageAttribute(image,key))) \
     member=(U8) strtol(attribute->value, (char **) NULL, 10); \
@@ -1011,7 +1045,7 @@ ModuleExport void UnregisterCINEONImage(void)
   const char \
     *definition_value; \
 \
-  if ((definition_value=AccessDefinition(image_info,"dpx",key+4))) \
+  if ((definition_value=AccessDefinition(image_info,"dpx",&key[4]))) \
     member=(U32) strtol(definition_value, (char **) NULL, 10); \
   else if ((attribute=GetImageAttribute(image,key))) \
     member=(U32) strtol(attribute->value, (char **) NULL, 10); \
@@ -1027,7 +1061,7 @@ ModuleExport void UnregisterCINEONImage(void)
   const char \
     *definition_value; \
 \
-  if ((definition_value=AccessDefinition(image_info,"dpx",key+4))) \
+  if ((definition_value=AccessDefinition(image_info,"dpx",&key[4]))) \
     member=(S32) strtol(definition_value, (char **) NULL, 10); \
   else if ((attribute=GetImageAttribute(image,key))) \
     member=(S32) strtol(attribute->value, (char **) NULL, 10); \
@@ -1043,7 +1077,7 @@ ModuleExport void UnregisterCINEONImage(void)
   const char \
     *definition_value; \
 \
-  if ((definition_value=AccessDefinition(image_info,"dpx",key+4))) \
+  if ((definition_value=AccessDefinition(image_info,"dpx",&key[4]))) \
     member.f=strtod(definition_value, (char **) NULL); \
   else if ((attribute=GetImageAttribute(image,key))) \
     member.f=strtod(attribute->value, (char **) NULL); \
@@ -1052,9 +1086,8 @@ ModuleExport void UnregisterCINEONImage(void)
 }
 
 /*
-  This macro uses strncpy on purpose.  The string is not required to
-  be null terminated, but any unused space should be filled with
-  nulls.
+  Header string values are not required to be null terminated, but any
+  unused space should be filled with nulls.
 */
 #define AttributeToString(image_info,image,key,member) \
 { \
@@ -1064,12 +1097,11 @@ ModuleExport void UnregisterCINEONImage(void)
   const char \
     *definition_value; \
 \
-  if ((definition_value=AccessDefinition(image_info,"dpx",key+4))) \
-    (void) strncpy(member,definition_value,sizeof(member)); \
-  else if ((attribute=GetImageAttribute(image,key))) \
-    (void) strncpy(member,attribute->value,sizeof(member)); \
-  else \
-    SET_UNDEFINED_ASCII(member); \
+  SET_UNDEFINED_ASCII(member); \
+  if ((definition_value=AccessDefinition(image_info,"dpx",&key[4]))) \
+    (void) memcpy(member,definition_value,Min(sizeof(member),strlen(definition_value))); \
+  else if ((attribute=GetImageAttribute(image,key)) && (attribute->value)) \
+    (void) memcpy(member,attribute->value,Min(sizeof(member),strlen(attribute->value))); \
 }
 
 static void GenerateCineonTimeStamp(char *date_str, size_t date_str_length, char*time_str, size_t time_str_length)
@@ -1127,7 +1159,7 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
 
   register const PixelPacket
     *p;
-  
+
   register long
     x;
 
@@ -1265,7 +1297,7 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
   cin_image_info.eol_pad = 0;
   /* End of channel padding */
   cin_image_info.eoc_pad = 0;
-  
+
   /*
     Image origination header.
   */
@@ -1330,7 +1362,7 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
   AttributeToString(image_info,image,"DPX:mp.frame.id",cin_mp_info.frame_id);
   /* Slate information */
   AttributeToString(image_info,image,"DPX:mp.slate.info",cin_mp_info.slate_info);
-  
+
   /*
     Open output image file.
   */
@@ -1433,7 +1465,7 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
     bits_per_sample = cin_image_info.channel_info[0].bits_per_sample;
     scale_from_short=(65535U / (65535U >> (16-bits_per_sample)));
 
-    scanline_bytes=image->columns*4;
+    scanline_bytes=MagickArraySize(image->columns,4);
     scanline=MagickAllocateMemory(unsigned char *,scanline_bytes);
     if (scanline == (unsigned char *) NULL)
       ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,image);
@@ -1444,9 +1476,9 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
         p=AcquireImagePixels(image,0,y,image->columns,1,&image->exception);
         if (p == (const PixelPacket *) NULL)
           break;
-        
+
         MagickBitStreamInitializeWrite(&bit_stream,scanline);
-        
+
         for (x=0; x < (long) image->columns; x++)
           {
             red   = ScaleQuantumToShort(p->red)/scale_from_short;
@@ -1474,7 +1506,7 @@ static unsigned int WriteCINEONImage(const ImageInfo *image_info,Image *image)
           if (QuantumTick(y,image->rows))
             if (!MagickMonitorFormatted(y,image->rows,&image->exception,
                                         SaveImageText,image->filename,
-					image->columns,image->rows))
+                                        image->columns,image->rows))
               break;
       }
     MagickFreeMemory(scanline);
